#-----------------------------------------------------------------------------
# Name:        setup.py
# Purpose:
#
# Author:      asumak
#
# Created:     17.06.2013
# Copyright:   (c) asumak 2012
# Licence:     <your licence>
#-----------------------------------------------------------------------------

from __future__ import print_function
from subprocess import Popen
import subprocess
import sys
import errno
import os

import_error_msg = "Nansat requires %s, which should be installed separately"

# Check if required packages are installed
try:
    import numpy
except ImportError:
    raise ImportError(import_error_msg % 'numpy')

try:
    from osgeo import gdal, osr, ogr
except ImportError:
    try:
        import gdal
        import osr
        import ogr
    except ImportError:
        raise ImportError(import_error_msg % 'gdal')

NAME                = 'nansat'
MAINTAINER          = "Nansat Developers"
MAINTAINER_EMAIL    = "nansat-dev@googlegroups.com"
DESCRIPTION         = "A scientist friendly Python toolbox for processing 2D satellite Earth observation data"
LONG_DESCRIPTION    = "A scientist friendly Python toolbox for processing 2D satellite Earth observation data"
URL                 = "https://github.com/nansencenter/nansat"
DOWNLOAD_URL        = "https://github.com/nansencenter/nansat"
LICENSE             = "GNU General Public License"
CLASSIFIERS         = [
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Atmospheric Science',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Utilities'
    ]
AUTHOR              = ("Anton Korosov, Morten W. Hansen, Kunt-Frode Dagestad, Aleksander Vines, Asuka Yamakawa")
AUTHOR_EMAIL        = "nansat-dev@googlegroups.com"
PLATFORMS           = ["Linux", "OS X", "Windows"]
MAJOR               = 1
MINOR               = 0
MICRO               = 18
ISRELEASED          = True
VERSION             = '%d.%d.%d' % (MAJOR, MINOR, MICRO) # Remember to remove "dev" when releasing
REQS                = [
                        "Pillow",
                        "pythesint",
                        "cfunits",
                        "urllib3",
                    ]

#----------------------------------------------------------------------------#
#                   Prepare compilation of C pixel functions
#----------------------------------------------------------------------------#
pixfun_module_name = '_pixfun_py{0}'.format(sys.version_info[0])
skip_compile = False
libraries = []
include_dirs = []
library_dirs = []
if os.name == 'nt':
    extra_compile_args = ['-nologo', '-DLL']
    try:
        path = os.environ['CONDA_PREFIX'] + '\Library\lib'
    except:
        # TODO: figure out how this works if you are not using conda
        path = os.environ['LIB'].split(';')
    #for iFolder in path:
    try:
        files = os.listdir(path)
    except:
        extra_link_args = []
    else:
        if 'gdal_i.lib' in files:
            extra_link_args = [path + '\gdal_i.lib']
else:
    extra_compile_args = ['-fPIC', '-Wall', '-Wno-long-long', '-pedantic', '-O3']
    extra_link_args = [] # not used currently


def _ask_gdal_config(resultlist, option, result_prefix):
    p = Popen(['gdal-config', option], stdout=subprocess.PIPE)
    t = p.stdout.read().decode().strip()
    if p.wait() != 0:
        return
    res = t.split()
    res = filter(lambda x: x.startswith(result_prefix), res)
    # '-I/usr/...' -> '/usr/...'
    res = [x[len(result_prefix):] for x in res]
    resultlist[:] = res


def use_gdal_config():
    _ask_gdal_config(include_dirs, '--cflags', '-I')
    _ask_gdal_config(library_dirs, '--libs', '-L')
    _ask_gdal_config(libraries,    '--libs', '-l')

def use_win_config():
    #try:
    path = os.environ['CONDA_PREFIX']
    include_dirs[:] = [path+'\Library\include']
    #except:
        # TODO: figure out how this works if you are not using conda
        # path = os.environ['LIB'].split(';')

try:
    if os.name == 'nt':
        use_win_config()
    else:
        use_gdal_config()
except Exception as e:
    if os.name == 'nt':
        print('WARNING: CONDA_PREFIX could not be found pixel functions will not be available.')
    else:
        print('WARNING: gdal-config could not be called, pixel functions will not be available.')
    print('Error details follow:')
    print(e)
    skip_compile = True

#----------------------------------------------------------------------------#
#                               Install package
#----------------------------------------------------------------------------#
from setuptools import setup, find_packages
from setuptools.command.install_scripts import install_scripts
from distutils import log
from distutils.extension import Extension
from distutils.errors import CCompilerError, DistutilsExecError,\
    DistutilsPlatformError

# Windows batch file handling
# from https://matthew-brett.github.io/pydagogue/installing_scripts.html
BAT_TEMPLATE = \
r"""@echo off
set mypath=%~dp0
set pyscript="%mypath%{FNAME}"
set /p line1=<%pyscript%
if "%line1:~0,2%" == "#!" (goto :goodstart)
echo First line of %pyscript% does not start with "#!"
exit /b 1
:goodstart
set py_exe=%line1:~2%
call %py_exe% %pyscript% %*
"""


class my_install_scripts(install_scripts):
    def run(self):
        install_scripts.run(self)
        if not os.name == "nt":
            return
        for filepath in self.get_outputs():
            # If we can find an executable name in the #! top line of the script
            # file, make .bat wrapper for script.
            with open(filepath, 'rt') as fobj:
                first_line = fobj.readline()
            if not (first_line.startswith('#!') and
                    'python' in first_line.lower()):
                log.info("No #!python executable found, skipping .bat "
                            "wrapper")
                continue
            pth, fname = os.path.split(filepath)
            froot, _ = os.path.splitext(fname)
            bat_file = os.path.join(pth, froot + '.bat')
            bat_contents = BAT_TEMPLATE.replace('{FNAME}', fname)
            log.info("Making %s wrapper for %s" % (bat_file, filepath))
            if self.dry_run:
                continue
            with open(bat_file, 'wt') as fobj:
                fobj.write(bat_contents)

# the following is adapted from simplejson's setup.py
if sys.platform == 'win32' and sys.version_info > (2, 6):
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    # It can also raise ValueError http://bugs.python.org/issue7511
    ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError,
                  IOError, ValueError)
else:
    ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)


def run_setup(skip_compile):
    if skip_compile:
        kw = dict()
    else:
        kw = dict(
            ext_modules = [
                Extension('{0}.{1}'.format(NAME, pixfun_module_name),
                          ['{0}/pixelfunctions/pixelfunctions.c'.format(NAME),
                           '{0}/pixelfunctions/{1}.c'.format(NAME, pixfun_module_name)],
                          include_dirs=include_dirs,
                          libraries=libraries,
                          library_dirs=library_dirs,
                          extra_compile_args=extra_compile_args,
                          extra_link_args=extra_link_args)
            ])


    # remove mapper_tests from installed packages
    packages = find_packages()
    #if 'mapper_tests' in packages:
    #    packages.remove('mapper_tests')

    setup(
        name=NAME,
        version=VERSION,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        url=URL,
        download_url=DOWNLOAD_URL,
        license=LICENSE,
        classifiers=CLASSIFIERS,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        platforms=PLATFORMS,
        packages=packages,
        package_data={NAME:["fonts/*.ttf", 'mappers/*.pl', 'tests/data/*.*']},
        scripts=[os.path.join('utilities', name) for name in
                    ['nansatinfo',
                     'nansat_add_coastline',
                     'nansat_geotiffimage',
                     'nansat_show',
                     'nansat_translate',
                     ]],
        cmdclass = {'install_scripts': my_install_scripts},
        install_requires=REQS,
        test_suite="nansat.tests",
        **kw
    )

try:
    run_setup(skip_compile)
except ext_errors:
    BUILD_EXT_WARNING = ("WARNING: The C extension could not be compiled, "
                         "pixel functions will not be available.")
    print('*' * 75)
    print(BUILD_EXT_WARNING)
    print("Failure information, if any, is above.")
    print("I'm retrying the build without the C extension now.")
    print('*' * 75)

    run_setup(True)

    print('*' * 75)
    print(BUILD_EXT_WARNING)
    print("Plain-Python installation succeeded.")
    print('*' * 75)



# Name:         opendap.py
# Purpose:      Abstract class Opendap is extended by some mappers
# Author:       Anton Korosov
# Licence:      This file is part of NANSAT. You can redistribute it or modify
#               under the terms of GNU General Public License, v.3
#               http://www.gnu.org/licenses/gpl-3.0.html

# http://cfconventions.org/wkt-proj-4.html

import os
import datetime
from dateutil.parser import parse
from time import sleep as time_sleep
import warnings
import numpy as np

try:
    from netCDF4 import Dataset
except ImportError:
    raise ImportError('''
         Cannot import Dataset from netCDF4.
         You cannot access OC CCI data but
         Nansat will work.''')

from nansat.nsr import NSR
from nansat.vrt import VRT
from nansat.tools import gdal

from nansat.exceptions import WrongMapperError


class Opendap(VRT):
    ''' Methods for all OpenDAP mappers '''

    P2S = {
        'H': 60*60,
        'D': 86400,
        'M': 30*24*60*60,
        'Y': 31536000,
        }

    ### TODOs:
    # add band metadata

    def test_mapper(self, filename):
        ''' Tests if filename fits mapper. May raise WrongMapperError '''
        baseURLmatch = False
        for baseURL in self.baseURLs:
            if filename.startswith(baseURL):
                baseURLmatch = True
                break
        if not baseURLmatch:
            raise WrongMapperError(filename)


    def get_dataset(self, ds):
        ''' Open Dataset '''
        if ds is None:
            try:
                ds = Dataset(self.filename)
            except:
                raise ValueError('Cannot open %s' % self.filename)
        elif type(ds) != Dataset:
            raise ValueError('Input ds is not netCDF.Dataset!')

        return ds

    def get_geospatial_variable_names(self):
        ''' Get names of variables with both spatial dimentions'''
        dsNames = []
        for var in self.ds.variables:
            if     (self.xName in self.ds.variables[var].dimensions and
                    self.yName in self.ds.variables[var].dimensions):
                dsNames.append(var)

        return dsNames

    def get_dataset_time(self):
        ''' Load data from time variable '''
        cachefile = ''
        if (type(self.cachedir) in [str, unicode] and
            os.path.isdir(self.cachedir)):
            # do caching
            cachefile = '%s_%s.npz' % (os.path.join(self.cachedir,
                                       os.path.split(self.filename)[1]),
                                       self.timeVarName)

        if os.path.exists(cachefile):
            dsTime = np.load(cachefile)[self.timeVarName]
        else:
            warnings.warn('Time consuming loading time from OpenDAP...')
            dsTime = self.ds.variables[self.timeVarName][:]
            warnings.warn('Loading time - OK!')

        if os.path.exists(cachefile):
            np.savez(cachefile, **{self.timeVarName: dsTime})

        return dsTime

    def get_layer_datetime(self, date, datetimes):
        ''' Get datetime of the matching layer and layer number '''

        if len(datetimes) == 1 or date is None:
            layerNumber = 0
        else:
            # find closest layer
            datetimeResolution = np.abs(datetimes[0] - datetimes[1])
            date = np.datetime64(date).astype('M8[s]')
            matchingDateDiff = np.min(np.abs(datetimes - date))
            if matchingDateDiff > datetimeResolution:
                raise ValueError('Date %s is out of range' % date)
            layerNumber = np.argmin(np.abs(datetimes - date))

        layerDate = datetimes[layerNumber]

        return layerNumber, layerDate

    def get_metaitem(self, url, varName, layerNo):
        ''' Set metadata for creating band VRT '''

        metaItem = {'src': {
                        'SourceFilename': '%s?%s.%s[%d][y][x]' % (url, varName, varName, layerNo),
                        'SourceBand': 1,
                            },
                    'dst': {
                        'name': varName,
                        'dataType': 6,
                            }
                    }

        for attr in self.ds.variables[varName].ncattrs():
            attrKey = attr.encode('ascii', 'ignore')
            attrVal = self.ds.variables[varName].getncattr(attr)
            if type(attrVal) in [str, unicode]:
                attrVal = attrVal.encode('ascii', 'ignore')
            if attrKey in ['scale', 'scale_factor']:
                metaItem['src']['ScaleRatio'] = attrVal
            elif attrKey in ['offset', 'add_offset']:
                metaItem['src']['ScaleOffset'] = attrVal
            else:
                metaItem['dst'][attrKey] = str(attrVal)

        return metaItem

    def create_vrt(self, filename, gdalDataset, gdalMetadata, date, ds, bands, cachedir):
        ''' Create VRT '''
        if date is None:
            warnings.warn('''
            Date is not specified! Will return the first layer.
            Please add date="YYYY-MM-DD"''')

        self.filename = filename
        self.cachedir = cachedir
        self.ds = self.get_dataset(ds)

        dsTime = self.get_dataset_time()

        dsDatetimes = self.convert_dstime_datetimes(dsTime)

        dsLayerNo, dsLayerDate = self.get_layer_datetime(date, dsDatetimes)

        if bands is None:
            dsVarNames = self.get_geospatial_variable_names()
        else:
            dsVarNames = bands

        # create VRT with correct lon/lat (geotransform)
        srcRasterXSize, srcRasterYSize = self.get_shape()
        srcGeoTransform = self.get_geotransform()
        self._init_from_dataset_params(srcRasterXSize, srcRasterYSize, srcGeoTransform, self.srcDSProjection)

        metaDict = [self.get_metaitem(filename, dsVarName, dsLayerNo)
                      for dsVarName in dsVarNames]

        self.create_bands(metaDict)

        # set time
        timeResSecs = self.get_time_coverage_resolution()
        self.dataset.SetMetadataItem('time_coverage_start', str(dsLayerDate))
        self.dataset.SetMetadataItem('time_coverage_end', str(dsLayerDate + timeResSecs))

    def get_time_coverage_resolution(self):
        ''' Try to fecth time_coverage_resolution and convert to seconds '''
        timeResSecs = 0
        if 'time_coverage_resolution' in self.ds.ncattrs():
            time_res = self.ds.time_coverage_resolution
            try:
                timeResSecs = int(time_res[1]) * self.P2S[time_res[2].upper()]
            except:
                warnings.warn('Cannot get time_coverage_resolution')

        return timeResSecs

    def get_shape(self):
        ''' Get srcRasterXSize and srcRasterYSize from OpenDAP '''
        return (self.ds.variables[self.xName].size,
                self.ds.variables[self.yName].size)


    def get_geotransform(self):
        ''' Get first two values of X,Y variables and create geoTranform '''

        xx = self.ds.variables[self.xName][0:2]
        yy = self.ds.variables[self.yName][0:2]
        return (xx[0], xx[1]-xx[0], 0, yy[0], 0, yy[1]-yy[0])
